#!/usr/bin/perl

#======================================================================
# CFILTL -- filter output of collectl
# 
# SYNOPSIS
# --------
# cfiltl allows you to extract individual values from collectl's
# plot-format output, possible manipulate them arithmetically, and
# present the chosen results in CSV format.
#
# cfiltl -f <collectl-file> [-a <num>] [-p] expr1 [expr2 [expr3 [...]]]
#
# collectl -p <collectl-file> -P --sep ^A | cfiltl [options as above]
#
# In the first form cfilt is passed a collectl *raw* file using the "-f"
# parameter. cfiltl then runs collectl as a child to process the raw
# file and produce plot-format output suitable for parsing by cfiltl.
#
# In the second form cfiltl reads plot-format output from STDIN.
# NOTE: the plot-format output that cfiltl reads from STDIN *must* have
# the columns delimited by CTRL-A (Control-A).
#
# OPTIONS
# -------
#  -f <collectl-file>, --file=<collectl-file>
#        specify collectl raw file to read using collectl
#
#  -p, --proconly
#        only process collectl records containing process data 
#
#  -a <num>, --average=<num>
#        calculate average values over <num> records before
#        evaluating tag-expresions
#
#  -s <char>, --sep=<char>
#        use <char> instead of space as CSV delimiter
#
#  --display[=<subsystem>]
#        extract subsystem and column names from <collectl-file> 
#        and display them. No <subsystem> means all.
#
#  --debug[=<num>]
#        enable debugging (max = 4, default = 1)
#
#  -v, --version
#        print version number and exit
#
# cfiltl will read from STDIN if the <collectl-file> is dash ("-"),
# or if the "-f" option is not present.
#
# DESCRIPTION
# -----------
# cfiltl implements a grammar for extracting individual data values and
# performing arithmetic on collectl's output. The grammar is specified
# in chunks called "subsystem-expressions".  cfiltl can be given an arbitrary
# number of subsystem-expressions. cfiltl's output consists of the output
# of each single subsystem-expression, in the order in which the
# subsystem-expressions were given on the command-line.
#
# In order to speed up processing, cfiltl generates Perl code based on
# the subsystem-expressions to extract only the collectl data needed to
# provide the requested results. This Perl code is compiled dynamically
# into subroutines. See "CFILTL INTERNALS" for more information.
# Among other things, this approach means that arithmetic in the
# tag-expressions is processed at the Perl native level.
#
# A subsystem-expression consists a subsystem-component, followed by an
# arbitrary number of "tag-expressions". The components are colon-delimited.
#
# The subsystem-component consists of:
#  1) a 3-character unique subsystem name (as found in collectl's plot output)
#  2) optionally either an "s" or a plus-sign ("+") meaning a summary subsystem
#     or summing is enabled
#  3) "selection-criteria" -- in the case of detail subsystems, optionally
#     an equals-sign ("=") followed by comma-separated names specifying
#     specific objects in the subsystem in question
#
#  Examples:
#    cpu      -- detail CPU subsystem
#    cpus     -- summary CPU subsystem
#    cpu=0,1  -- only values for the CPUs 0 and 1
#    cpu+=0,1 -- summed values over CPUs 0 and 1
#    dsk+     -- summed values for disk detail subsystem
#    nets     -- summary network subsystem
#
# A tag-expression can be a combination of the following:
#  1) tags (collectl column-names), from collectl's plot-format output
#     (with some restrictions -- see below)
#  2) functions such as int(), sin(), cos(), log(), sqrt(), norm(), etc
#  3) operators: +, -, *, /, **, %
#  4) parenthesis to group sub-expresions
#  5) literal numerical values, such as 99, 3.14159, etc
#
# Tag naming
#   tag-names are taken from collectl's plot-format column-name headers with
#   the following restrictions/exceptions:
#   - all column-names are converted to lower-case
#   - if a column-name contains a arithmetic operator (%*/+-)
#     (examples: "Wait%", "Intrpt/sec")
#     the operator and everything following it is removed.
#   - a few column-names contain dashes: "L-Avg1", "L-Avg5", "L-Avg15".
#     These are converted to "ldavg1", "ldavg5", and "ldavg15".
#
#   Thus:
#       "Wait%"      becomes "wait"
#       "Intrpt/sec" becomes "intrpt"
#       "proc/sec"   becomes "proc"
#
#   Otherwise, tags must correspond 1-to-1 to column-names. Thus,
#   "proc" and "procque" are two different tags.
#
#   Use "--display[=<subsys>]" to view tag names for one or all subsystems
#   that are available in a collectl raw file.
#  
#     
# Tags can be considered variables. The actual values are extracted
# from each collectl record and then the tag-expression are evaluated.
# There are several things that affect tags:
#  1) "normalization"
#     normalization is intended to allow plotting multiple curves
#     on the same graph for data that falls in wildly differing ranges,
#     for example cpu load percentages (user, sys, idle, etc) with
#     context-switches or disk throughput/IOs. Thus it allows
#     forcing data occuring in an arbitrarily high range into a specified
#     range.
#
#     There are two normalization notations:
#       1) appending a hash ("#") to a tag
#       2) using the pseudo-function "norm(<expression>[,<max>])"
#
#	  appending hash to a tag
#     -----------------------
#     If a hash ("#") is placed immediately following
#     a tag, the value will be normalized to a range between 0 and 100.
#     If an whole number immediately follows the hash, then the value
#     will be normalized to a range between 0 and the number given.
#	  examples:
#       user     -- substitute the value in the "user" column (no normalization)
#       ctx#     -- normalize the value of context-switches to 0-100
#       ctx#200  -- normalize the value of context-switches to 0-200
#
#     This can only be used on a single tag.
#     This form of  normalization can only be performed on
#     the transient values corresponding to  a tag, and not to a larger part,
#     or the entire tag-expression. Thus, "int(intrpt#/8)" will first
#     normalize the value extracted for "intrpt" (for some or all CPUs,
#     depending on the subsystem), and then divide the normalized value
#     by 8, and then truncate the result to an integer.
#     This short-form is intended to be a short-cut for trivial cases. For
#     non-trivial expressions, see below.
#
#
#     Pseudo-function "norm(...)"
#     ---------------------------
#     This can contain an arbitrary arithmetic expression, including
#     an arbitrary number of tags. It *cannot* be nested.
#     Examples:
#       norm(txkb+rxkb)     -- normalize the sum of txkb and rxkb to range 0-100
#       norm(txkb+rxkb,200) -- normalize the sum of txkb and rxkb to range 0-200
#       norm((txkb+rxkb)/2) -- normalize (txkb + rxkb)/2 to range 0-100
#
#     Because normalization requires knowing the maximum value over all
#     collectl records, using normalization means cfiltl must process
#     the entire output of collectl before it can produce results.
#     In order to do this, it caches the individual values that it needs
#     in order to calculate a value for a tag-expression. It does *not*
#     cache all collectl output.
#
#  2) summing
#     if a plus-sign follows the subsystem name in the subsystem-component,
#     summing is enabled for all tag-expressions in that particular
#     subsystem-expression. "summing" means:
#     1) if no selection-criteria are present, the value used to replace
#        the tag is the sum over all objects in the subsystem
#     2) if selection-criteria are present, the value is the sum
#        over all specified objects
#
#   3) averaging
#     cfiltl can calculate averages (arithmetic mean) over an
#     arbitrary number of samples before any further processing is performed.
# 
# Order of steps to evaluate an expression:
# 1) extract individual values, or sum, as required, from each record
# 2) if averaging is desired, cache these intermediate values and
#    produce an average over N collectl records
# 3) if no normalization is requested, evaluate tag-expressions and
#    print a CSV line of values
# 4) if normalization is requested, calculate normalized values for
#    each tag, and then evaluate each tag-expression.
#
# EXAMPLES of subsystem-expressions:
#
#  cpus:user:sys:idle:wait
#    CSV OUTPUT: 4 columns: cpus:user, cpus:sys, cpus:idle, cpus:wait
#    extract "user", "sys", "idle", "wait" values from cpu subsystem
#
#  cpus:user+nice:sys:intrpt#
#    CSV OUTPUT: 3 columns: cpus:user+nice, cpus:sys, and cpus:intrpt(normalized to 0-100)
#  
#  cpu+=0,1:user:sys
#    CSV OUTPUT: 2 columns: cpu+=0,1:user, cpu+=0,1:sys
#    the sum of 'user' for CPU0 + CPU1, the sum of 'sys' for CPU0 + CPU1
#    I.e., for each record: <cpu-0-user + cpu-1-user>, <cpu-0-sys + cpu-1-sys>
#
#  cpu=0,1:user:sys
#    CSV OUTPUT: 4 columns: cpu=0:user, cpu=0:sys, cpu=1:user, cpu=1:sys
#    extract the individual user and sys values for CPUs 0 and 1
#
#  cpu+=0,1,2,3:intrpt/4
#    CSV OUTPUT: 1 column: cpu+=0,1,2,3:intrpt/4
#    sum the intrpt value for CPUs 0-3 and then divide by 4, in other
#    words, the average per-CPU interrupt value (assuming 4-core system)
#
#  net=eth0:norm(rxkb+txkb):norm(rxpkt+txpkt)
#    CSV OUTPUT: 2 columns: net=eth0:rxkb+txkb#100, net=eth0:rxpkt+txpkt#100
#
# Avoiding division-by-zero errors
#   cfiltl tries to determine if a division-by-zero error for a particular
#   tag-expression is possible (for example, when the denominator is a tag).
#   In such cases, it uses Perl's exception-handling mechanism to trap
#   such div-by-zero errors, and substitutes zero as the result.
#
# CFILTL INTERNALS
# ----------------
# First, the subsystem-expressions are parsed, during which Perl-code is
# generated to extract values from each record and perform
# arithmetic to arrive at a result to be output.
# Up to three functions are dynamically generated by build_code():
#  - now_func()       - always run on each record
#  - later_func()     - run on averaged samples, or at end of record processing
#  - avg_high_func()  - run on averaged samples to save maximum values
# 
# Each collectl record is parsed and stored in "$record" in MAIN.
# $record is passed to process_record().
#
# If normalization and averaging are disabled, all necessary steps are
# performed in "now_func()", which is called from process_record().
#
# If normalization or averaging is enabled, "now_func()" only extracts
# values, which get cached in an array referred to as "$sample". When
# normalization is enabled, but not averaging, "now_func()" also contains
# code to save maximum values used later in normalizing.
#
# If averaging is enabled, but not normalization, "later_func()" is called
# for each averaged batch of samples to evaluate the subsystem-expressions
# and generate CSV output.
#
# If normalization is enabled, but not averaging, "now_func()" stores
# temporary values in $sample, which is cached locally in process_records()
# until all records have been processed for which an average is desired.
# 
# If averaging AND normalization are enabled, "now_func()" stores
# averaged values in $sample. "avg_high_func()" is called on each averaged
# sample to save maximum values for later normalization. Then the averaged
# sample gets cached in the global variable
# "@SAMPLES" until all collectl records have been processed, at which
# point "later_func()" is called on each sample.
#
# A "sample" is an array of values, in which the first element is the
# timestamp from a collectl record. All following elements are used for
# holding transient values extracted from collectl records corresponding
# to the tags used in the tag-expressions. The sample elements may, but
# must not necessarily, correspond to to the elements in cfiltl's output.
# This is because a tag-expression, which always corresponds to one or more
# output values, can contain zero, one, or many tags.
#
#
# Author: Rob Urban <urban@unix-beratung.de>
# Date:   June, 2014
#======================================================================

#======================================================================
# CHANGES
#
# 1.2 -> 1.3
#	- help function added
#	- --display=<subsystem> added to parse collectl raw file and display
#		available tags for one, or all subsystems
#	- based on the assumption that only process data is displayed separately
#		from all other subsystems, cfiltl extracts the collection header
#		using collectl and checks whether process data is present.
#		If yes, it waits until it sees the process column description header
#		and then checks if the subsystem-expressions refer to unknown
#		subsystems.
#	- cfiltl now expects CTRL-A delimited plot format output on STDIN
#		if no collectl-file is explicitly provided using "-f"
#
# 1.1 -> 1.2
#	- when averaging, take timestamp from middle sample
#
# 1.0 -> 1.1
#	- added max values to header column descriptions for normalized tags
#	- increased efficiency slightly
#======================================================================

use strict;

use Getopt::Long;
use IO::Handle;

#==============================================================
# globals
#==============================================================
my $VERSION				= '1.4';
my $IN_FILE;						# input file, may be "-"
my $DEBUG				= 0;
my %FIELD_COUNT;
my %COL_OFFSET;
my @MAX_VALS;
my @NORM_HIGH;
my @CUR_OUTPUT;
my $CUR_OUT_POS			= 0;
my @RESET_VALUES;
my @SAMPLES;
#my $CUR_SAMPLE;
my $SAMPLE_POS			= 1;	# starts at 1 to leave room for timestamp
my @SAMPLE_TO_OUTPUT;
my $IMMEDIATE			= 1;
my @NORMALIZATION;
my @SUBS;
my @AFTER_SUM_SUBS;
my $AVERAGING			= 0;
my @PRE_SUM_STORAGE;
my @OUT_HEADER;
my $COL_SEP_CHAR		= ' ';
my $PROCONLY			= 0;
my $AVERAGE				= 0;
my @AVERAGES;
my %FOUND_TAGS;
my %USED_TAGS;
my %METADATA;
my $READ_STDIN			= 0;
my @FUNCTIONS			= qw(cos sin tan sqrt log exp abs atan2 int norm);
my $SHOW_SS;

# %SS_MAP is currently wrong, and not used for much. Still trying to
# figure out how to determine when a non-existent subsystem has been
# specified in a subsystem-expression

my %SS_MAP = (
	b	=> { name => 'bud', type => 'summary', desc => 'buddy info', },
	c	=> { name => 'cpu', type => 'summary', desc => 'CPU', },
	d	=> { name => 'dsk', type => 'summary', desc => 'Disk', },
	f	=> { name => 'nfs', type => 'summary', desc => 'NFS V3', },
	i	=> { name => 'ino', type => 'summary', desc => 'Inode and Filesystem', },
	j	=> { name => 'int', type => 'summary', desc => 'Interrupt', },
	l	=> { name => 'lus', type => 'summary', desc => 'Lustre', },
	m	=> { name => 'mem', type => 'summary', desc => 'Memory', },
	n	=> { name => 'net', type => 'summary', desc => 'Network', },
	s	=> { name => 'soc', type => 'summary', desc => 'Sockets', },
	t	=> { name => 'tcp', type => 'summary', desc => 'TCP', },
	x	=> { name => 'ict', type => 'summary', desc => 'Interconnect', },
	y	=> { name => 'sla', type => 'summary', desc => 'Slabs', },
	C	=> { name => 'cpu', type => 'detail', desc => 'CPU', },
	D	=> { name => 'dsk', type => 'detail', desc => 'Disk', },
	E	=> { name => 'env', type => 'detail', desc => 'Environmental', },
	F	=> { name => 'nfs', type => 'detail', desc => 'NFS Data', },
	J	=> { name => 'int', type => 'detail', desc => 'Interrupts', },
	L	=> { name => 'lus', type => 'detail', desc => 'Lustre', },
	M	=> { name => 'num', type => 'detail', desc => 'Memory Node (NUMA)', },
	N	=> { name => 'net', type => 'detail', desc => 'Network', },
	T	=> { name => 'tcp', type => 'detail', desc => 'TCP Counters', },
	X	=> { name => 'ict', type => 'detail', desc => 'Interconnect', },
	Y	=> { name => 'sla', type => 'detail', desc => 'Slabs', },
	Z	=> { name => 'pro', type => 'detail', desc => 'Processes', },
);
my %SS_MAP_REV;

#-------------------------------------------------------------------
# indices into $record array for 'general' (everything but process),
# and 'process' statistics. I use an array and constants for speed.
#-------------------------------------------------------------------
use constant GEN_IND => 0;
use constant PROC_IND => 1;

#==============================================================
# main
#==============================================================
my $res = GetOptions(
	'file=s'		=> \$IN_FILE,
	'debug:1'		=> \$DEBUG,
	'separator=s'	=> \$COL_SEP_CHAR,
	'proconly'		=> \$PROCONLY,
	'average=i'		=> \$AVERAGE,
	'version'		=> sub { print "Version $VERSION\n"; exit; },
	'help'			=> \&usage,
	'display-ss:s'	=> \$SHOW_SS,
);
$res || print "failed to process arguments successfully.\n" && usage();

#--------------------------------------------------------------
# create subsystem info structure with 3-char subsystem name key
#--------------------------------------------------------------
foreach my $ss_char (keys(%SS_MAP)) {
	$SS_MAP_REV{ $SS_MAP{$ss_char}->{name} } = {
		char	=> $ss_char,
		type	=> $SS_MAP{$ss_char}->{type},
		desc	=> $SS_MAP{$ss_char}->{desc},
	};
}

#--------------------------------------------------------------
# parse extraction expressions and generate extraction/evaluation functions
#--------------------------------------------------------------

# first determine if any normalization has been requested. If not,
# set $IMMEDIATE to 1
my $expr_count = 0;
foreach my $expr (@ARGV) {
	if ($expr =~ /(norm\(|#)/) {
		$IMMEDIATE = 0;
		$DEBUG && print "setting IMMEDIATE = 0\n";
	}
}

my $now_code = "sub now_func {\nmy (\$record,\$sample) = \@_;\nmy \$val;\n";
my $later_code	= "sub later_func {\nmy \$sample = shift;\n";
my $avg_code;

foreach my $expr (@ARGV) {
	my ($nc, $lc, $ac) = parse_expression($expr, $expr_count);
	$now_code	.= $nc;
	$later_code .= $lc;
	$avg_code	.= $ac;
	$expr_count++;
}
$now_code .= '}';
$later_code .= '}';
$avg_code = "sub avg_high_func {\nmy \$sample = shift;\n$avg_code}";

@RESET_VALUES = (0) x $CUR_OUT_POS;

if ($DEBUG > 2) {
	print "\nNOW CODE:\n$now_code\n--END--\n\nLATER_CODE:\n$later_code\n--END--\n\nAVG_CODE:\n$avg_code\n--END--\n";
	print "OUT_HEADER: @OUT_HEADER\n";
	print "RESET_VALUES=(",join(',', @RESET_VALUES),")\n";
	$DEBUG > 3 && exit;
}
eval $now_code;
defined(&now_func) || die "failed to compile now_func().\n";
eval $later_code;
defined(&later_func) || die "failed to compile later_func().\n";
eval $avg_code;
defined(&avg_high_func) || die "failed to compile avg_func().\n";

my $line_num = 0;
my $header_lines_seen = 0;
my $stat_type;
my $prev_timestamp;
my $record = [];
my @length_to_index_map;
my @length_to_type_map;
my $record_counter = 0;

#--------------------------------------------------------------
# print header line
#--------------------------------------------------------------
$IMMEDIATE && !defined($SHOW_SS) && print_line('#timestamp', \@OUT_HEADER);

#--------------------------------------------------------------
# handle input-file
#--------------------------------------------------------------
if (!defined($IN_FILE) || ($IN_FILE eq '-')) {
	$READ_STDIN = 1;
}
elsif (defined($IN_FILE) && ! -f $IN_FILE) {
	die "file [$IN_FILE] not found\n";
}

#--------------------------------------------------------------
# get some interesting information from the collectl header
#--------------------------------------------------------------
$READ_STDIN || get_collectl_metadata();

#--------------------------------------------------------------
# run collectl to read file
#--------------------------------------------------------------
my $fh = IO::Handle->new;
if (!$READ_STDIN) {
	my $pid = open($fh, '-|', 'collectl', '-p', $IN_FILE, '-P', '--sep', "\cA");
	defined($pid) || die "failed to run collectl\n";
}
else {
	open($fh, '>&STDIN') || die "failed to dup STDIN";
}

while(<$fh>) {
	chomp;
	my @fields = split(/\cA/, $_);
	if (@fields == 1) {
		die qq[you forgot to use CTRL-A as field separator ("--sep ^A")\n];
	}
	if ($DEBUG > 1) {
		s/\cA/ /g;
		print "LINE> $_\n";
	}
	my $atoms = @fields;
	if (/^#/) {
		#-----------------------------------------------
		# column names
		#-----------------------------------------------
		if ($header_lines_seen == 0) {
			$stat_type = 'gen';
			$length_to_index_map[$atoms] = GEN_IND;
			$length_to_type_map[$atoms] = 'gen';
		}
		elsif ($fields[2] eq 'PID') {
			$stat_type = 'proc';
			$length_to_index_map[$atoms] = PROC_IND;
			$length_to_type_map[$atoms] = 'proc';
		}
		else {
			die "failed to recognize type of headers: [$_]\n";
		}
		$FIELD_COUNT{$stat_type} = $atoms;
		parse_header_line($stat_type, $line_num, \@fields);
		$header_lines_seen++;
		if (!$READ_STDIN) {
			if ($METADATA{subsystems} !~ /Z/ || $stat_type eq 'proc') {
				delayed_checks();
			}
		}
	}
	else {
		#-----------------------------------------------
		# data values
		#-----------------------------------------------
		my $ts = join(' ', @fields[0,1]);
		if ($ts ne $prev_timestamp) {
			if ($DEBUG > 1) {
				print "## TIMESTAMP DIFFERS (ts=[$ts], prev=[$prev_timestamp] ##\n";
				if (exists($record->[PROC_IND])) {
					print "PROC DATA PRESENT\n";
				}
			}
			if (!$PROCONLY || exists($record->[PROC_IND])) {
				@$record && process_record($prev_timestamp, $record);
				$record_counter++;
			}
			$DEBUG > 1 && print "saving prev_timestamp...\n";
			$prev_timestamp = $ts;
		}
		# collect new input data in $record
		# maybe do this better later
		if ($length_to_index_map[$atoms] == GEN_IND) {
			$DEBUG  > 1 && print "--> GEN <--\n";
			#$record->[GEN_IND] = [ @fields[2..$#fields] ];
			$record->[GEN_IND] = \@fields;
			if ($DEBUG > 2) {
				my $count = 0;
				foreach my $tmp (@{ $record->[GEN_IND] }) {
					printf("  [%4d] %8s", $count, $tmp);
					$count++;
				}
				print "\n";
			}
		}
		else {
			$DEBUG > 1 && print "--> PROC <--\n";
			#$record->[PROC_IND]->{$fields[2]} = [ @fields[2..$#fields] ];
			$record->[PROC_IND]->{$fields[2]} = \@fields;
		}
	}
}
continue {
	$line_num++;
}

# process last record (possibly)
if (!$PROCONLY || exists($record->[PROC_IND])) {
	process_record($prev_timestamp, $record);
}

$READ_STDIN || $fh->close;

$IMMEDIATE && exit;

#--------------------------------------------------------------
# substitute max values for place holders in the header labels
#--------------------------------------------------------------
for(my $samp_pos=0; $samp_pos<@MAX_VALS; $samp_pos++) {
	#print "samp_pos=$samp_pos, max=".(defined($MAX_VALS[$samp_pos])?$MAX_VALS[$samp_pos]:'undef')."\n";
	defined($MAX_VALS[$samp_pos]) || next;
	my $out_pos = $SAMPLE_TO_OUTPUT[$samp_pos];
	my ($this, $that) = ('\\['.$samp_pos.'\\]', $MAX_VALS[$samp_pos]);
	#print "fixing OUT_HEADER[$out_pos], replace '$this' with '$that'\n";
	$OUT_HEADER[$out_pos] =~ s/$this/$that/;
}

print_line('#timestamp', \@OUT_HEADER);

foreach my $sample (@SAMPLES) {
	later_func($sample);
	print_line($sample->[0], \@CUR_OUTPUT);
	@CUR_OUTPUT = @RESET_VALUES;
}

#==============================================================
# subs
#==============================================================

sub print_line
{
	my ($ts, $valref) = @_;
	print $ts, $COL_SEP_CHAR, join($COL_SEP_CHAR, @{$valref}), "\n";
}

{
	# the following two variables are only visible to process_record()
	my $in_counter = 0;
	my @avgs;

	sub process_record
	{
		my ($timestamp, $record) = @_;

		my $sample;

		$DEBUG && print "** process_record(), ts=[$timestamp], in_counter=$in_counter **\n";

		defined($SHOW_SS) && return;
		#-----------------------------------------------------
		# call extraction function
		#-----------------------------------------------------
		my $ts = conv_timestamp($timestamp);
		$sample = [$ts];
		now_func($record, $sample);
		@$record = ();

		my $avg_ref;
		my $tmp;
		$in_counter++;

		if ($AVERAGE) {
			if ($in_counter % $AVERAGE == 0) {
				$DEBUG > 1 && print "  averaging...\n";
				push(@avgs, $sample);
				$sample = calc_averages(\@avgs);
				@avgs = ();
				if ($IMMEDIATE) {
					later_func($sample);
					print_line($sample->[0], \@CUR_OUTPUT);
					@CUR_OUTPUT = @RESET_VALUES;
				}
				else {
					avg_high_func($sample);
					push(@SAMPLES, $sample);
				}
			}
			else {
				push(@avgs, $sample);
			}
			return;
		}

		if ($IMMEDIATE) {
			print_line($ts, \@CUR_OUTPUT);
			@CUR_OUTPUT = @RESET_VALUES;
		}
		else {
			push(@SAMPLES, $sample);
		}
	}
}

#-------------------------------------------------------------------
# calc_averages() reckons the arithmetic mean for each index
# in an array of samples
#-------------------------------------------------------------------
sub calc_averages
{
	my $aref = shift;

	my $out;
	# how many samples do I have?
	my $num_samples = @{ $aref };

	# get the index of the middle sample from which to get the timestamp
	my $ts_idx = int($num_samples/2);
	#print "n_samples=$num_samples, ts_idx=$ts_idx\n";
	$out->[0] = $aref->[$ts_idx]->[0];

	my $len = @{ $aref->[0] };
	my $sum;

	for(my $i=1; $i<$len; $i++) {
		$sum = 0;
		foreach my $s (@{ $aref }) {
			$sum += $s->[$i];
		}
		$out->[$i] = $sum/$num_samples;
	}

	return $out;
}

#-------------------------------------------------------------------
# conv_timestamp()
# input:
#  date and time in the form "YYYYMMDD HH:MM:SS"
#
# output:
#  string in form "YYYY:MM:DD:HH:MM:SS"
#-------------------------------------------------------------------
sub conv_timestamp
{
	my $ts = shift;

	my ($year, $mon, $day, $h, $m, $s) = $ts =~ /^(\d{4})(\d{2})(\d{2}) (\d+):(\d+):(\d+)$/;
	defined($year) || die "failed to parse timestamp [$ts]";

	return sprintf('%s:%s:%s:%s:%s:%s', $year, $mon, $day, $h, $m, $s);
}

#-------------------------------------------------------------------
# parse_header_line()
# input:
#  a collectl column description line
#
# modifies %COL_OFFSET and %FOUND_TAGS
#-------------------------------------------------------------------
sub parse_header_line
{
	my ($stat_type, $line_num, $aref) = @_;

	$DEBUG > 1 && print "** parse_header_line(), stat_type=[$stat_type] **\n";
	shift(@$aref);	# get rid of #Date
	shift(@$aref);	# get rid of Time
	my $count = 2;
	foreach my $field (@$aref) {
		$field = lc($field);
		$field =~ s!l-avg!ldavg!i;
		$field =~ s![%/*-+].*$!!;
		if ($stat_type eq 'proc') {
			$COL_OFFSET{pro}->{$field} = $count;
			$FOUND_TAGS{pro}->{detail}->{$field} = 1;
			$DEBUG > 2 && print "ss=[$stat_type], stat=[$field], pos=$count\n";
		}
		else {
			my ($ss, $item, $stat);
			($ss, $item, $stat) = $field =~ /^\[(?:([^:]{3})[^:]*)(?::([^\]]+))?\](.*)$/;
			defined($ss) || die "failed to parse atom [$field]";
			$DEBUG > 2 && print "ss=[$ss], item=[$item], stat=[$stat], pos=$count\n";
			$item = 'summ' if (!defined($item));
			$COL_OFFSET{$ss}->{$item}->{$stat} = $count;
			$FOUND_TAGS{$ss}->{($item eq 'summ')?'summary':'detail'}->{$stat} = 1;
		}
		$count++;
	}
	check_tags();
}

sub check_tags
{
	my @ss_not_found;

	foreach my $ss (keys(%USED_TAGS)) {
		if (!exists($FOUND_TAGS{$ss})) {
			push(@ss_not_found, $ss);
			next;
		}
		foreach my $type (keys(%{ $USED_TAGS{$ss} })) {
			if (!exists($FOUND_TAGS{$ss}->{$type})) {
				die "no $type data present for SS [$ss]\n";
			}
			foreach my $tag (keys(%{ $USED_TAGS{$ss}->{$type} })) {
				if (!exists($FOUND_TAGS{$ss}->{$type}->{$tag})) {
					die "no column [$tag], type [$type] data present for SS [$ss]\n";
				}
			}
		}
	}

	return @ss_not_found;
}

sub found_all_headers
{
	$DEBUG > 1 && print "found_all_headers()\n";

	exists($METADATA{subsystems}) || die "cannot find subsystems in meta-data\n";
	foreach my $ss_char (split(//, $METADATA{subsystems})) {
		$DEBUG > 1 && print "  checking [$ss_char]\n";
		my $ss = $SS_MAP{$ss_char}->{name};
		my $type = $SS_MAP{$ss_char}->{type};
		if (!exists($FOUND_TAGS{$ss}) || !exists($FOUND_TAGS{$ss}->{$type})) {
			$DEBUG > 1 && print "    did not find [$ss]\n";
			return 0;
		}
		$DEBUG > 1 && print "    found [$ss]\n";
	}
	return 1;
}

#-------------------------------------------------------------------
# parse_expression()
# parses a subsystem-expression into subsystem component and
# tag-expresssions. Then calls build_code() for each tag-expresion
#
# rearrangement of result columns
# -------------------------------
# because parse_expression() iterates over each tag-expr, but build_code()
# iterates over each object selected, if there are multiple objects in the
# selection-criteria, the order of code produced is (example: 3 objects,
# 3 tag-exprs):
#	obj1/tag-expr1 obj2/tag-expr1 obj3/tag-expr1
#   obj1/tag-expr2 obj2/tag-expr2 obj3/tag-expr2
#   obj1/tag-expr3 obj2/tag-expr3 obj3/tag-expr3
#
# this seems less intuitive than the order:
#	obj1/tag-expr1 obj1/tag-expr2 obj1/tag-expr3
#	obj2/tag-expr1 obj2/tag-expr2 obj2/tag-expr3
#	obj3/tag-expr1 obj3/tag-expr2 obj3/tag-expr3
#
# thus, assuming the current output column number (OCN) is N,
# parse_expression() calls build_code() with OCN=N, OCN=N+1, OCN=N+2,
# and the increment (the number of tag-expressions) that build_code()
# needs to add to the OCN for each pass, if it needs multiple passes.
# Thus build_code() would use the following OCNs for the above example:
#	OCN		expr
#   ----	---------------
#	N		obj1/tag-expr1
#	N+3		obj2/tag-expr1
#	N+6		obj3/tag-expr1
#
# and on the next pass, for tag-expr2:
#	N+1		obj1/tag-expr2
#	N+4		obj2/tag-expr2
#	N+7		obj3/tag-expr2
#
# and on the next pass, for tag-expr3:
#	N+2		obj1/tag-expr3
#	N+5		obj2/tag-expr3
#	N+8		obj3/tag-expr3
#
# and if we sort by OCN we have:
#	N		obj1/tag-expr1
#	N+1		obj1/tag-expr2
#	N+2		obj1/tag-expr3
#	N+3		obj2/tag-expr1
#	N+4		obj2/tag-expr2
#	N+5		obj2/tag-expr3
#	N+6		obj3/tag-expr1
#	N+7		obj3/tag-expr2
#	N+8		obj3/tag-expr3
#
# Before moving on to the next ss-expr, parse_expression() increments
# the current OCN position by:
#	NumOfItems * NumOfTagExprs
#
# In the above case, 3 * 3, so by 9
#-------------------------------------------------------------------
sub parse_expression
{
	my $expr = lc(shift);
	my $count = shift;

	my ($now_code, $later_code, $avg_code);

	$DEBUG > 1 && print "parse_expression()\n";

	my @parts = split(/:/, $expr);

	#---------------------------------------------------
	# deal with subsystem
	#---------------------------------------------------
	my $ss_part = shift(@parts);
	my ($ss, $summ, $plus, $selection) = $ss_part =~ /^(.{3})(s)?(\+?)(?:=(.*))?$/;
	$DEBUG > 1 && print "  ss=[$ss]", ($summ?", summ=1":''), ($plus?", plus=1":''), ($selection?", selection=$selection":''), "\n";
	# check whether subsystem exists
	exists($SS_MAP_REV{$ss}) || die "subsystem [$ss] unknown\n";

	my @items;
	my $summing = 0;

	if (length($summ)) {
		if (defined($selection)) {
			die "cannot provide selection-criteria for summary subsystem [$ss]\n";
		}
		@items = ('summ');
	}
	if (length($plus)) {
		$summ && die "summing on a summary subsystem makes no sense\n";
		$summing = 1;
	}
	if (length($selection)) {
		@items = split(/,/, $selection);
	}
	elsif (!length($summ)) {
		# summing always enabled for detail subsystems if no selection-criteria
		# present
		$summing = 1;
	}

	#---------------------------------------------------
	# construct bodies of subs
	#---------------------------------------------------
	my $tmp_pos = $CUR_OUT_POS;
	my $incr = @parts;
	my ($nc, $lc, $ac);
	my $jump;	# number of output columns build_code() has occupied

	# insert process-subsystem specific check for presence of data
	if ($ss eq 'pro') {
		$now_code = insert_pro_check_begin($summing, $tmp_pos, $incr, @items);
	}

	foreach my $tag_expr (@parts) {
		($nc, $lc, $ac, $jump) = build_code($ss, $summ, $summing, $tag_expr, $tmp_pos, $incr, @items);
		$now_code	.= $nc;
		$later_code .= $lc;
		$avg_code	.= $ac;
		$tmp_pos++;
	}
	if ($ss eq 'pro') {
		$now_code .= "}\n";
	}
	$CUR_OUT_POS += $jump * $incr;
	return ($now_code, $later_code, $avg_code);
}

sub untokenize
{
	my @tokens = @_;

	my @out;
	foreach my $token (@tokens) {
		push(@out, $token->{val});
	}

	return join(' ', @out);
}

#-------------------------------------------------------------------
# tokenize()
# breaks up a tag-expression into its component parts:
#  function-names
#  tags
#  operators and parentheses
#  literal numerical values
#
# If old-style normalization notation is used ("tag#[<max>]"), tokenize()
# will generate new-style tokens "norm", "(", "tag", ",", "<max>", ")" and
# add them to the list.
#
# returns some information about tag-expression and an array of
# "tokens" (components) found. Array elements corresponding to
# tags are references to a hash with the key "tag" containing the
# tag-name, and possible "max" if normalization has been requested
# for that tag
#
# In the meantime, tokenize() also keeps track of when it's in the
# middle of a "norm(...)" subexpression, and does not worry about
# division within such expressions, because that is handled later
# by need_norm_eval() which is called from gen_col_expression().
#-------------------------------------------------------------------
sub tokenize
{
	my ($ss, $str) = @_;

	$DEBUG > 4 && print "tokenize(), ss=[$ss], str=[$str]\n";

	my $normalize = 0;
	my @tokens;
	my $found_div = 0;
	my @after_div;
	my $use_eval = 0;
	my $norming = 0;
	my $depth = 0;
	my $debug = 0;

	while(length($str)) {
		$debug && print "tokenize-top: fdiv=$found_div, norming=$norming, depth=$depth\n";
		if ($str =~ /^([a-z0-9]{3,})\((.*)$/) {
			# function-name
			my ($func, $rest) = ($1, $2);
			$DEBUG > 4 && print "  function-name [$func]\n";
			check_func($func);
			push(@tokens, { type => 'func', val => $func });
			$str = '('.$rest;
			$found_div && push(@after_div, 'func');
			if ($func eq 'norm') {
				$norming = 1;
			}
			#$depth++;
		}
		elsif ($str =~ /^([a-zA-Z]\w{2,})(?:(#)([\d.]+)?)?(.*)$/) {
			# tag
			my ($tag, $hash, $max, $rest) = ($1, $2, $3, $4);
			$DEBUG > 4 && print "  tag [$tag]\n";
			my $ref = { type => 'tag', val => $tag, label => "REPLACE($tag)" };
			if (defined($hash)) {
				$ref->{max} = defined($max) ? $max : 100;
				$DEBUG > 4 && print "    max: $ref->{max}\n";
				$normalize = 1;
				my @tmp = (
					{ type => 'func', val => 'norm' },
					{ type => 'op', val => '(' },
					$ref,
					defined($max) ? (
							{ type => 'op', val => ',' },
							{ type => 'val', val => $max },
						) : (),
					{ type => 'op', val => ')' },
				);
				push(@tokens, @tmp);
			}
			else {
				push(@tokens, $ref);
			}
			$found_div && push(@after_div, 'tag');
			$str = $rest;
		}
		elsif ($str =~ m!^([()+/%,-]|\*\*?)(.*)$!) {
			# parenthesis, comma or arithmetic operator
			my ($op, $rest) = ($1, $2);
			$str = $rest;
			$DEBUG > 4 && print "  paren-or-op [$op]\n";
			push(@tokens, { type => 'op', val => $op });
			$found_div && push(@after_div, 'op');
			if ($op eq '(') { $depth++; }
			elsif ($op eq ')') {
				$depth--;
				if ($norming && !$depth) {
					$norming = 0;
				}
			}
			elsif (!$norming && $op eq '/') { $found_div++; }
		}
		elsif ($str =~ /^([\d.]+)(.*)$/) {
			# literal integer or float value
			my ($val, $rest) = ($1, $2);
			$DEBUG > 4 && print "  value [$val]\n";
			push(@tokens, { type => 'val', val => $val });
			$str = $rest;
			$found_div && push(@after_div, 'val');
		}
	}
	$debug && print "after_div: (", join(',', @after_div), ")\n";
	if ($found_div) {
		# cover trivial case where dividing by a constant
		if (@after_div == 1 && $after_div[0] eq 'val') {
			$use_eval = 0;
		}
		elsif (@after_div > 1 && $after_div[0] eq 'val' && $after_div[1] eq 'op') {
			# or where denomitor a constant followed by (presumably) a paren
			$use_eval = 0;
		}
		else {
			$use_eval = 1;
		}
	}
	$debug && print "tokening returns use_eval=$use_eval\n";
	return ($normalize, $found_div, $use_eval, @tokens);
}


#-------------------------------------------------------------------
# insert_pro_check_begin()
#
# inserts code into now_func() to substitute zero values for
# any results requested for processes when process-data is not
# present in a record
#-------------------------------------------------------------------
sub insert_pro_check_begin
{
	my ($summing, $cur_pos, $num_expr, @items) = @_;

	my $now_code;

	my $s_pos = $SAMPLE_POS;
	my $count = $summing ? $num_expr : $num_expr * @items;

	$now_code = 'if (!exists($record->[' . PROC_IND . '])) {'."\n";
	for(my $i=0; $i<$count; $i++) {
		if ($IMMEDIATE && !$AVERAGE) {
			$now_code .= "  \$CUR_OUTPUT[$cur_pos] = 0;\n";
		}
		else {
			$now_code .= "  \$sample->[$s_pos] = 0;\n";
		}
		$cur_pos++;
		$s_pos++;
	}
	$now_code .= "} else {\n";

	return $now_code;
}

#-------------------------------------------------------------------
# build_code()
# a bit hairy...
# constructs now_func(), later_func(), and avg_high_func() based
# on tokens found in a tag-expression
#-------------------------------------------------------------------
sub build_code
{
	my ($ss, $summary, $summing, $tag_expr, $cur_pos, $incr, @items) = @_;

	my ($now_code, $later_code);
	my $jump = 0;
	my $col_expr;
	my $avg_code;
	my $trivial;
	my $use_sprintf = 0;
	my $label;

	my ($norm, $found_div, $use_eval, @tokens) = tokenize($ss, $tag_expr);
	$DEBUG && print "build_code(): found_div=$found_div, use_eval=$use_eval, tokens: ",untokenize(@tokens),"\n";

	if ($summing) {
		($now_code, $col_expr, $avg_code, $trivial, $label)
			= gen_col_expression($ss, $summing, \@tokens, \@items, $cur_pos);
		$use_sprintf = $found_div || !$trivial;
		if ($IMMEDIATE && !$AVERAGE) {
			$now_code .= check_eval("\$CUR_OUTPUT[$cur_pos]", $col_expr, $use_eval, $use_sprintf);
		}
		else {
			$later_code .= check_eval("\$CUR_OUTPUT[$cur_pos]", $col_expr, $use_eval, $use_sprintf);
		}
		my $tmp = "${ss}+";
		if (@items) { $tmp .= '='.join(',', @items); }
		#$OUT_HEADER[$cur_pos] = "$tmp:$tag_expr";
		$OUT_HEADER[$cur_pos] = "$tmp:$label";
		$cur_pos += $incr;
		$jump++;
	}
	else {
		#-----------------------------------------------------
		# covers detail subsystems without summing, or
		# summary subsystems
		#-----------------------------------------------------
		foreach my $item (@items) {
			$DEBUG && print "ITEM: [$item]\n";
			$col_expr = '';
			my ($nc, $lc, $ac);
			($nc, $col_expr, $ac, $trivial, $label)
				= gen_col_expression($ss, $summing, \@tokens, [$item], $cur_pos);
			$now_code	.= $nc;
			#$later_code .= $lc;
			$avg_code	.= $ac;
			$use_sprintf = $found_div || !$trivial;
			if ($IMMEDIATE && !$AVERAGE) {
				$now_code .= check_eval("\$CUR_OUTPUT[$cur_pos]", $col_expr, $use_eval, $use_sprintf);
			}
			else {
				$later_code .= check_eval("\$CUR_OUTPUT[$cur_pos]", $col_expr, $use_eval, $use_sprintf);
			}
			my ($ss_tmp, $items);
			if ($summary) {
				$ss_tmp = "${ss}s";
			}
			else {
				$ss_tmp = $ss . "=$item";
			}
			#$OUT_HEADER[$cur_pos] = "$ss_tmp:$tag_expr";
			$OUT_HEADER[$cur_pos] = "$ss_tmp:$label";
			$cur_pos += $incr;
			$jump++;
		}
	}
	return ($now_code, $later_code, $avg_code, $jump);
}

#-------------------------------------------------------------------
# gen_col_expresion()
#
# This is where the music is playing...
#
# It is a hideous state-machine, with states defined by:
#	 $norming, $depth, $found_comma
#
# does two things:
# 1) builds a perl column-expression to return a result corresponding
#    to a tag-expression
# 2) produces code for now_func(), and avg_high_func()
#    if necessary
# the column-expression is returned, and build_code() decides
# whether it should be added to now_func() or later_func()
#-------------------------------------------------------------------
sub gen_col_expression
{
	my ($ss, $summing, $tokens_ref, $items_ref, $cur_pos) = @_;

	my ($target, $now_code, $col_expr, $avg_code);
	my $trivial = 1;
	my $label = '';
	my $norming = 0;
	my $norm_expr = '';
	my $depth = 0;
	my $found_comma = 0;
	my $norm_high = 100;
	my @norm_tokens;

	foreach my $token (@{$tokens_ref}) {
		$DEBUG && print "norming=$norming, depth=$depth, token-type=$token->{type}, token-val=[$token->{val}], label=[$label]\n";
		if ($norming) {
			if ($found_comma) {
				if ($token->{type} ne 'val') {
					die "no norm-high found after comma\n";
				}
				$norm_high = $token->{val};
				$DEBUG && print "setting norm-high to $norm_high\n";
				$found_comma = 0;
				next;
			}
			elsif (!$depth) {
				# token must be "(", so toss
				$depth++;
				next;
			}
			elsif ($token->{val} eq ',') {
				$found_comma = 1;
				next;
			}
			elsif ($token->{val} eq '(') {
				$depth++;
				$norm_expr .= $token->{val};
				$label .= $token->{val};
				push(@norm_tokens, $token);
				next;
			}
			elsif ($token->{val} eq ')') {
				$depth--;
				if (!$depth) {
					#------------------------------------------------
					# end of norm(...)
					#------------------------------------------------
					my $norm_eval = need_norm_eval(@norm_tokens);
					$label .= ')#'. $norm_high;
					my $lhs = "\$sample->[$SAMPLE_POS] = \$val";
					$now_code .= check_eval($lhs, $norm_expr, $norm_eval, 0);
					if ($AVERAGE) {
						$avg_code .= ''
						."\$MAX_VALS[$SAMPLE_POS] = \$sample->[$SAMPLE_POS] "
						."if (\$sample->[$SAMPLE_POS] > \$MAX_VALS[$SAMPLE_POS]);\n";
					}
					else {
						$now_code .= "\$MAX_VALS[$SAMPLE_POS] = \$val if (\$val > \$MAX_VALS[$SAMPLE_POS]);\n";
					}
					$NORM_HIGH[$SAMPLE_POS] = $norm_high;
					$col_expr .= "calc_norm($SAMPLE_POS, \$sample)";
					$trivial = 0;
					$DEBUG && print "adding max to label\n";
					$label .= "(max=[$SAMPLE_POS])";
					$found_comma = 0;
					$norming = 0;
					$norm_high = 100;
					$SAMPLE_TO_OUTPUT[$SAMPLE_POS] = $cur_pos;
					$SAMPLE_POS++;
				}
				else {
					#------------------------------------------------
					# inside a parenthetical expression in norm(...)
					#------------------------------------------------
					$norm_expr .= $token->{val};
					$label .= $token->{val};
					push(@norm_tokens, $token);
				}
				next;
			}
		}

		if ($token->{type} eq 'tag') {
			$label .= $token->{val};
			if ($summing) {
				$target = "sum_items('$ss', '$token->{val}', \$record, " . lit_list($items_ref) . ")";
			}
			else {
				# construct term to access single datum
				$target = rec_access_code($ss, $items_ref->[0], $token->{val});
			}

			if ($norming) {
				$norm_expr .= $target;
			}
			else {
				if ($IMMEDIATE && !$AVERAGE) {
					$col_expr .= $target;
				}
				else {
					# *this* expression doesn't have normalization, but
					# somewhere else it's needed, so everything must
					# go into @CUR_SAMPLE
					$now_code .= "\$sample->[$SAMPLE_POS] = $target;\n";
					$col_expr .= "\$sample->[$SAMPLE_POS]";
				}
				$SAMPLE_TO_OUTPUT[$SAMPLE_POS] = $cur_pos;
				$SAMPLE_POS++;
			}

			# note subsystem, item, and column-name for later sanity check
			if ($summing) {
				$USED_TAGS{$ss}->{detail}->{$token->{val}} = 1;
			}
			elsif ($items_ref->[0] eq 'summ') {
				$USED_TAGS{$ss}->{summary}->{$token->{val}} = 1;
			}
			else {
				$USED_TAGS{$ss}->{detail}->{$token->{val}} = 1;
			}
			push(@norm_tokens, $token);
		}
		elsif ($token->{val} =~ /^norm/) {
			die "no nested normalization allowed\n" if $norming;
			$norming = 1;
			$label .= 'norm(';
			$norm_expr = '';
			@norm_tokens = ();
		}
		elsif ($token->{val} eq ',') {
			die "found comma outside norm(...)\n";
		}
		elsif ($norming) {
			$norm_expr .= $token->{val};
			$label .= $token->{val};
			push(@norm_tokens, $token);
		}
		else {
			$col_expr .= $token->{val};
			$label .= $token->{val};
		}
	}

	#print "returning label=[$label]\n";
	return ($now_code, $col_expr, $avg_code, $trivial, $label);
}

sub need_norm_eval
{
	my @tokens = @_;

	my @denom;

	my $after_div = 0;
	my $debug  = 0;

	$debug && print "need_norm_eval(), expr=[".untokenize(@tokens)."\n";

	foreach my $token (@tokens) {
		$debug && print "  token: [$token->{val}], after_div=$after_div\n";
		if ($after_div) {
			if ($token->{val} eq ')') {
				denom_unsafe(@denom) && return 1;
				$after_div = 0;
				@denom = ();
			}
			elsif ($token->{val} eq '/') {
				denom_unsafe(@denom) && return 1;
				@denom = ();
			}
			else {
				push(@denom, $token);
			}
		}
		elsif ($token->{type} eq 'op' && $token->{val} eq '/') {
			$after_div = 1;
		}
	}
	@denom && denom_unsafe(@denom) && return 1;
	return 0;
}

sub denom_unsafe
{
	my @tokens = @_;

	foreach my $token (@tokens) {
		if ($token->{type} eq 'tag' || $token->{type} eq 'func') {
			return 1;
		}
	}
	return 0;
}

sub check_eval
{
	my ($lhs, $rhs, $use_eval, $use_sprintf) = @_;

	if ($use_sprintf) {
		$rhs = qq[sprintf('%.2f', $rhs)];
	}

	$use_eval || return "$lhs = $rhs;\n";

	return "eval { $lhs = $rhs; }; if (\$@) { $lhs = 0; }\n";
}

sub rec_access_code
{
	my ($ss, $item, $tag) = @_;

	my $pos;
	if ($ss eq 'pro') {
		$pos = "\$COL_OFFSET{pro}->{$tag}";
		return '$record->[' . PROC_IND . "]->{$item}->[$pos]";
	}
	else {
		$pos = "\$COL_OFFSET{$ss}->{$item}->{$tag}";
		return '$record->[' . GEN_IND . "]->[$pos]";
	}
}

sub calc_norm
{
	my ($pos, $sample) = @_;
	my $tmp;
	eval {
		#$tmp = sprintf('%.2f', $sample->[$pos] / $MAX_VALS[$pos] * $NORM_HIGH[$pos]);
		$tmp = $sample->[$pos] / $MAX_VALS[$pos] * $NORM_HIGH[$pos];
	};
	if ($@) {
		$tmp = 0;
	}
	return $tmp;
}

sub lit_list
{
	#my @list = @_;
	my $list_ref = shift;

	@$list_ref || return '';

	return "'" . join("','", @$list_ref) . "'";
}

sub check_func
{
	my $name = shift;

	my $re = '(' . join('|', @FUNCTIONS) . ')';
	if ($name !~ /^$re$/)
	{
		die "function [$name] unknown\n";
	}
}

#-------------------------------------------------------------------
# sum_items()
# sums the values for specified tag for all, or specified, items in a subsystem
#-------------------------------------------------------------------
sub sum_items
{
	my ($ss, $tag, $record, @items) = @_;

	my $sum = 0;

	my $pos;

	if ($ss eq 'pro') {
		$pos = $COL_OFFSET{pro}->{$tag};
		foreach my $pid (@items ? @items : keys(%{ $record->[PROC_IND] })) {
			# TODO: figure out a way to check that this exists once
			$sum += $record->[PROC_IND]->{$pid}->[$pos];
		}
	}
	else {
		foreach my $name (@items ? @items : keys(%{ $COL_OFFSET{$ss} })) {
			next if ($name eq 'summ');
			$pos = $COL_OFFSET{$ss}->{$name}->{$tag};
			if ($DEBUG > 1) {
				my $tmp = $record->[GEN_IND]->[$pos];
				print "sum_items: adding $tmp to $sum (name=[$name])\n";
			}
			$sum += $record->[GEN_IND]->[$pos];
		}
	}

	return $sum;
}

#-------------------------------------------------------------------
# get_collectl_metadata()
# calls collectl on data-file with --showheader
# to get header, which it parses for interesting information
#-------------------------------------------------------------------
sub get_collectl_metadata
{
	my $fh = IO::Handle->new;
	my $pid = open($fh, '-|', 'collectl', '-p', $IN_FILE, '--showheader');
	defined($pid) || die "failed to run collectl\n";

	while(<$fh>) {
		chomp;
		s/\s*$//;
		if (/^#\s+subsys:\s+(\S+)\s+options:\s+(\S.*)$/i) {
			$METADATA{subsystems} = $1;
			my @rest = split(' ', $2);
			while(my $key = shift(@rest)) {
				my $val = shift(@rest);
				if ($key eq 'Interval:') {
					$METADATA{intervals} = [ split(/:/, $val) ];
				}
			}
		}
	}
	$fh->close;
}

sub usage
{
	print <<_EOF_;
usage:

  cfiltl -f <collectl-file> [-a <num>] [-p] expr1 [expr2 [expr3 [...]]]

  collectl -p <collectl-file> -P --sep ^A | cfiltl [options as above]

OPTIONS
-------
  -f <collectl-file>, --file=<collectl-file>
        specify collectl raw file to read using collectl

  -p, --proconly
        only process collectl records containing process data 

  -a <num>, --average=<num>
        calculate average values over <num> records before
        evaluating tag-expresions

  -s <char>, --sep=<char>
        use <char> instead of space as CSV delimiter

  --display[=<subsystem>]
        extract subsystem and column names from <collectl-file> 
        and display them. No <subsystem> means all.

  --debug[=<num>]
        enable debugging (max = 4, default = 1)

  -v, --version
        print version number and exit

cfiltl will read from STDIN if the <collectl-file> is dash ("-"),
or if the "-f" option is not present.
_EOF_
	print "\nFunctions Supported: ", join(', ', @FUNCTIONS), "\n";
	exit;
}

#-------------------------------------------------------------------
# delayed_checks() is called after cfiltl has (hopefully) seen
# all column description header lines.
#-------------------------------------------------------------------
sub delayed_checks
{
	#-------------------------------------------------------------------
	# show tag names for selected subsystem, or all subsystems
	#-------------------------------------------------------------------
	if (defined($SHOW_SS)) {
		my @list;
		if (length($SHOW_SS)) {
			exists($FOUND_TAGS{$SHOW_SS}) || die "no subsystem [$SHOW_SS] found\n";
			@list = ($SHOW_SS);
		}
		else {
			@list = keys(%FOUND_TAGS);
			exists($FOUND_TAGS{pro}) && push(@list, 'pro');
		}
		printf("[%s - %7s] -- TAG LIST --\n", 'SUB', 'TYPE');
		foreach my $ss (@list) {
			foreach my $type (qw(summary detail)) {
				exists($FOUND_TAGS{$ss}->{$type}) || next;
				printf("[%s - %7s] ", $ss, $type);
				print join(', ', sort keys(%{ $FOUND_TAGS{$ss}->{$type} })), "\n";
			}
		}
		exit 0;
	}

	#-------------------------------------------------------------------
	# check if ss-expressions contain unknown subsystem names
	#-------------------------------------------------------------------
	$DEBUG > 1 && print "checking for absent ss data...\n";
	my @ss_not_found = check_tags();
	if (@ss_not_found) {
		die "no data present for subsystems: ".join(',', @ss_not_found)."\n";
	}
}
